(function (root, factory) {
    root.requestHandler = factory(root, root._, root.jQuery);
}(typeof window !== 'undefined' ? window : this, function (global, _, $) {
    /**
     * GLOBAL Variables - pre-defined in window
     * @property {string} GLOBAL_SERVER_IP - server ip
     * @property {string} GLOBAL_SECRET - client_secret
     * @property {string} checkInTurnTvFrom - 0|1
     * @property {string} checkInTurnTvTo - 0|1
     */
    /**
     * localStorage Variables - localStorage.getItem(property)
     * @property {string} access_token
     * @property {string} refresh_token
     * @property {string} client_id - serial_number
     * @property {string} room_number - room_number
     */
    var DEFAULT_HEADERS = {
        'Content-Type': 'application/json',
    };
    /*
        TODO : after reqQueue pop if fail do again
        type of requests
        request (type, url, header, param, callbackFn)
            requestAM (a.k.a fetchAM) :9443/asyncManager
                requestQMS (--- N/A ---) wrap-up function
                requestSSE (--- N/A ---) wrap-up function
            requestAPI :60080/api/v2/etc
                requestMonitoring
        request only work in initialization

        functions
            getAsyncManagerUri
            renewAccessToken
            fetchAMData
            getAMData
     */
    var initialized = false;
    var reqQueue = [];
    var REQUEST_TYPES = {
        TICKER: 'TICKER',
        ONE_CLICK_SERVICE: 'ONE_CLICK_SERVICE',
        NOTICE: 'NOTICE',
        QMS: 'QMS',
        PMS: 'PMS',
        NOTIFIER: 'NOTIFIER',
        AD_BANNER: 'AD_BANNER',
        QMS_SERVICE_LIST: 'GETTVQMSSERVICE',
        QMS_NEW_WORK_ORDER: 'SETTVQMSREQUEST',
        QMS_JOB_HISTORY: 'GETTVQMSREQUEST',
        QMS_REFRESH: 'QMSREFRESH',
        IOT_BUTTON: 'IOT_BUTTON',
        QMS_TV_SERVICE: 'QMSTVSERVICE',
        QMS_TV_REQUEST: 'QMSTVREQUEST',
        QMS_UPDATE_REQUEST: 'UPDATETVQMSREQUEST'
    };
    var DEBUG = false;
    var subscriptions = {};
    var pageSubscriptions = {};

    function log(message) {
        if(DEBUG){
            log(message);
        }
    }

    function publish(type, data) {
        type = type.toUpperCase();
        _.forEach(subscriptions[type], function (cb) {
            try {
                cb(data);
            } catch (e) {
            }
        });
        _.forEach(pageSubscriptions[type], function (cb) {
            try {
                cb(data);
            } catch (e) {
            }
        });
    }

    function getRequestURI(port, uri) {
        if(global.GLOBAL_SERVER_IP) {
            return ['https://', global.GLOBAL_SERVER_IP, ':', port, uri].join('');
        } else {
            return false;
        }
    }

    function subscribe(type, cb, fromPage) {
        if (fromPage) {
            pageSubscriptions[type.toUpperCase()].push(cb);
        } else {
            subscriptions[type.toUpperCase()].push(cb);
        }
    }

    function unsubscribeAll() {
        _.forEach(REQUEST_TYPES, function (v, k) {
            subscriptions[v] = [];
        });
        _.forEach(REQUEST_TYPES, function (v, k) {
            pageSubscriptions[v] = [];
        });
    }

    function unsubscribeFromPage() {
        _.forEach(REQUEST_TYPES, function (v, k) {
            pageSubscriptions[v] = [];
        });
    }

    function _request(type, url, headers, params, callback) {
        console.log('--- requestHandler._request : '+type+' / '+url+' [START]---');
        if (!initialized) {
            reqQueue.push([type, url, headers, params, callback]);
            console.log('--- requestHandler._request : '+type+' / '+url+' [NOT INITIALIZED]---');
            return Promise.reject('NOT INITIALIZED');
        }
        _.extend(headers, DEFAULT_HEADERS);
        return new Promise(function(resolve, reject) {
            $.ajax({
                type: type,
                url: url,
                headers: headers,
                data: JSON.stringify(params),
                dataType: 'json',
                cache: false
            }).done(function (response) {
                resolve(response);
                //console.log(JSON.stringify(response));
                console.log('--- requestHandler._request : '+type+' / '+url+' [DONE]---');
            }).fail(function (response, code) {
                reject(response);
                console.log(response, code);
                console.log('--- requestHandler._request : '+type+' / '+url+' [FAIL]---');
            }).error(function (jqXHR, textStatus, errorThrown) {
                reject();
                console.log(textStatus);
                console.log('--- requestHandler._request : '+type+' / '+url+' [ERROR]---');
            });
        }).then(function(response) {
            try {
                callback(response);
                return response;
            } catch (e) {
                console.log('_request callback error');
            }
        });;
    }

    function requestAM(action, md5, params, callback) {
        return new Promise(function (resolve, fail) {
            if (localStorage.getItem('access_token') == null) {
                if (!initialized) {
                    var type = 'POST';
                    var url = getRequestURI('9443', '/asyncManager');
                    var amHeaders = {
                        Authorization: null,
                        action: action,
                        md5: md5
                    };
                    var amParams = _.extend({}, params);
        
                    reqQueue.push([type, url, amHeaders, amParams, callback]);
                    console.log('--- requestHandler._request : '+type+' / '+url+' [NOT INITIALIZED]---');
                    return Promise.reject('NOT INITIALIZED');
                } else {
                    refreshAccessToken()
                    .then(function (response) {
                        global.console.warn('--- requestHandler.requestAM Response ' + response.access_token + ' ---');
                        localStorage.setItem('access_token', response.access_token);
                        localStorage.setItem('refresh_token', response.refresh_token);
                        return resolve(response.access_token);
                    });
                }
            } else {
                return resolve(localStorage.getItem('access_token'));
            }
        }).then(function (token) {
            var type = 'POST';
            var url = getRequestURI('9443', '/asyncManager');
            var amHeaders = {
                Authorization: 'Bearer ' + token,
                action: action,
                md5: md5
            };
            var amParams = _.extend({}, params);
            return _request(type, url, amHeaders, amParams, callback || _.noop);
        });
    }

    function refreshAccessToken(cb) {
        global.console.warn('--- requestHandler.refreshAccessToken ---');
        var params = {
            client_id: localStorage.getItem('client_id'),
            client_secret: GLOBAL_SECRET,
            refresh_token: localStorage.getItem('refresh_token'),
            grant_type: 'refresh_token'
        }
        return requestAPI('POST', '/api/v2/auth/tokens',params, cb || _.noop);
    }

    function requestAPI(type, url, params, callback) {
        var apiUrl = getRequestURI('60080', url);
        var apiHeaders = {
            Authorization: 'Bearer ' + global.localStorage.getItem('access_token')
        };
        var apiParams = _.extend({}, params);
        return _request(type, apiUrl, apiHeaders, apiParams || {}, callback || _.noop);
    }

    function getEvents(events) {
        /* asyncManager:9443 action SSE */
        var reqEvents = (_.isUndefined(events) || !_.isArray(events)) ? _.keys(REQUEST_TYPES) : events;
        var eventOptions = {};
        if(_.includes(events, 'notifier')){
            eventOptions.last_notice_id = (typeof (Storage) !== 'undefined') ? Number(localStorage.getItem('notifier_event_id')) : 0;
            eventOptions.pcn_last_notice_id = (typeof (Storage) !== 'undefined') ? localStorage.getItem('pcn_notifier_event_id') : 0;
        }
        var params = {
            params: {
                serialNo: global.localStorage.getItem('client_id'),
                events: reqEvents,
                eventOptions: eventOptions
            }
        };
        var callbackFn = function (response) {
            var resEvents = JSON.parse(response);
            _.forEach(resEvents, function (data, reqType) {
                log('*********reqType' + reqType);
                log('**** data ' + JSON.stringify(data));
                publish(reqType.toUpperCase(), data);
            });
        };
        requestAM('sse', '', params, callbackFn);
    }

    function getEventMsg() {
        var params = {
            type: 'eventMsg'
        };
        var callbackFn = function(response) {
            if(response.status === 'success' && $.isPlainObject(response.data) && response.data.id !== localStorage.getItem('lastEventMsgId')) {
                log('--- requestHandler.getEventMsg : '+response.data.insertAt+' / '+response.data.id+' ---');
                global.localStorage.setItem('lastEventMsgId', response.data.id);
                var events = JSON.parse(data.param || '[]');
                getEvents(events);
            }
        };
        requestAM('getPushMessage','', params, callbackFn);
    }

    function requestQMS(action, params) {
        requestAM(action, '', params, function (response) {
            publish(action, response);
        });
    }

    function init() {
        initialized = true;
        /*
                accessToken = global.localStorage.getItem('access_token');
                roomId = global.localStorage.getItem('room_number');
                serialNo = global.localStorage.getItem("client_id");
        */
        processRequestQueue(reqQueue);
    }

    function processRequestQueue(queue) {
        console.log('called processRequestQueue');
        if(queue.length) {
            var argument = queue.pop();
            argument[2]['Authorization']='Bearer ' + global.localStorage.getItem('access_token');
            _request(argument[0], argument[1], argument[2], argument[3], argument[4])
                .finally(function() {
                    processRequestQueue(queue);
                });
        }
    }

    unsubscribeAll();
    return {
        init: init,
        request: _request,
        requestAPI: requestAPI,
        requestAM: requestAM,
        getEvents: getEvents,
        requestQMS: requestQMS,
        getEventMsg: getEventMsg,
        publish: publish,
        subscribe: subscribe,
        unsubscribeAll: unsubscribeAll,
        unsubscribeFromPage: unsubscribeFromPage,
        subscriptions: subscriptions,
        pageSubscriptions: pageSubscriptions,
        refreshAccessToken: refreshAccessToken
    };
}));
