/* Summary: A library meant to manage the authorization for those smart apps that requires authentication  */
(function (root, factory) {
    root.smartAppAuthHandler = factory(root, root._, root.requestHandler, root.Hashids, root.CryptoJS, root.moment);
}(typeof window !== 'undefined' ? window : this, function (global, _, requestHandler, Hashids, CryptoJS, moment) {
    'use strict';

    function getAuthToken(app) {
        //Do not try to get auth tokens if TV has not internet connection
        if (localStorage.getItem('tv_ip_address') == '0.0.0.0' ||
            localStorage.getItem('tv_ip_address') == null ||
            localStorage.getItem('access_token') == null ||
            app == null
        ) {
            return Promise.resolve(false);
        } else {
            return fetchToken(app)
                .then(function (token) {
                    if (!token.error) {
                        localStorage.setItem(app + '_token', JSON.stringify(token));
                        if (app.toLowerCase() === 'guestflix') {
                            return getSignedGuestFlixToken();
                        } else if (app.toLowerCase() === 'showtime') {
                            showtimeHashRoom();
                            return token;
                        } else {
                            return token;
                        }
                        // localStorage.removeItem('getting_tokens');
                    } else {
                        return false;
                    }
                }, function (err) {
                    throw new Error(JSON.stringify(err));
                });
        }

    }

    function fetchToken(app) {
        var appToken = JSON.parse(localStorage.getItem(app + '_token'));
        // logConsole("@@@ SMARTAPP AUTH STORED " +localStorage.getItem(app + '_token'));
        if (appToken != null && !appToken.error) {
            if (isTokenValid(app)) {
                return Promise.resolve(appToken);
            } else {
                logConsole(" Token expired fetching new token from AM");
                return requestHandler.requestAM('smart_apps_token', '', {
                    app: app
                });
            }
        } else {
            logConsole("Token not found in localstorage get it from AM");
            return requestHandler.requestAM('smart_apps_token', '', {
                app: app
            });
        }
    }
    /**** Utility functions ****/
    function isTokenValid(app) {
        var token = JSON.parse(localStorage.getItem(app + '_token'));
        var isTokenValid = false;
        switch (app) {
            case 'hbo':
            case 'showtime':
                var nowUtc = moment().utc();
                var lastUpd = moment.utc(token.last_update);
                var diffSec = nowUtc.diff(lastUpd, 'seconds');
                var expires_in = (typeof token.expires_in != 'undefined' && token.expires_in != null) ? token.expires_in : 43200; //12hr showtime
                isTokenValid = (diffSec <= parseInt(expires_in));
                break;

            case 'guestflix':
                if (pickGuestFlixSecret() != null) {
                    isTokenValid = true;
                }
                break;
        }
        return isTokenValid;
    }

    function pickGuestFlixSecret() {
        var auth = JSON.parse(localStorage.getItem('guestflix_token'));
        localStorage.setItem('guestflix_siteId', auth.siteId.toString());
        var nowEpoch = moment().valueOf();
        var secret = null;
        auth.secrets.forEach(function (tokenObj) {
            if (nowEpoch > tokenObj.starts && nowEpoch < tokenObj.expires) {
                secret = tokenObj;
            }
        });
        // localStorage.setItem('guestflix_token', JSON.stringify(token));
        return secret;
    }

    function showtimeHashRoom() {
        var room = localStorage.getItem('room_number');
        var salt = 'SHOWTIME_SALT_HASH';
        var alphabet = '1234567890abcdefghijklmnopqrxyz';
        var chars = room.split('');

        for (var x = 0; x < chars.length; x++) {
            if (isNaN(parseInt(chars[x], 10))) {
                chars[x] = chars[x].charCodeAt(0);
            }
        }
        var numeric_string = chars.join('');

        var room_int = parseInt(numeric_string);
        var hashids = new Hashids(salt, 16, alphabet);
        var roomHashId = hashids.encode(room_int);
        localStorage.setItem("room_hash", roomHashId);

        return roomHashId;
    }

    function getSignedGuestFlixToken() {
        var roomNum = localStorage.getItem('room_number');
        var modelName = localStorage.getItem('model_name');
        var serialNum = localStorage.getItem('client_id');
        var expires_in = moment().unix() + 43200;
        var starts_in = moment().unix();
        var siteId = localStorage.getItem('guestflix_siteId');

        var header = {
            "alg": "HS256",
            "typ": "JWT"
        };
        var data = {
            "sub": "AT",
            "aud": "GFLIX",
            "rmn": roomNum.toString(),
            "iss": "lgprocentric",
            "dmd": modelName,
            "exp": expires_in,
            "dmf": "LG",
            "iat": starts_in,
            "did": serialNum,
            "sid": siteId,
            "rml": "Bedroom"
        };
        /*For QA purposes the room # must be 101
        var roomNum = 101; & "sid": "783",
        "env": "qa",*/

        //Enable QA Mode if siteId eq to 783
        if (siteId == 783) {
            data.rmn = "101";
            data.env = "qa";
        }

        var wordArrayHeader = CryptoJS.enc.Utf8.parse(JSON.stringify(header));
        var encodedHeader = CryptoJS.enc.Base64.stringify(wordArrayHeader);

        var wordArrayData = CryptoJS.enc.Utf8.parse(JSON.stringify(data));
        var encodedData = CryptoJS.enc.Base64.stringify(wordArrayData);

        var accessToken = encodedHeader + "." + encodedData;
        var secret = pickGuestFlixSecret();

        if (_.isObject(secret)) {
            var signature = CryptoJS.HmacSHA256(accessToken, secret.secret);
            signature = CryptoJS.enc.Base64.stringify(signature);
            var signedToken = accessToken + '.' + signature;
        }

        return {
            'accessToken': signedToken,
            'expires_in': expires_in,
            'starts_in': starts_in
        }
    }

    return {
        getAuthToken: getAuthToken,
        showtimeHashRoom: showtimeHashRoom
    }
}));
